#! /usr/bin/python

import numpy
import random

max_iter = 100000
energy_max = 200
starting_temp = 20.0

stall_threshold = 10
temp_increment = 5

def tikz_show(g, (a, b), temp, dt, k):
    print "~"
    # print "\\begin{figure}[htbp]"
    print "\\begin{tikzpicture}[scale=0.5, every node/.style={scale=0.8}]"
    for n in g.nodes:
        color="blue!10"
        if n == a or n == b: color="red!10"
        print "\\node[rectangle, minimum width=4em, minimum height=4em, fill=%s, draw] (%d) at (%d, %d) {%d, weight: %d};"  % (color, n.id, n.x, n.y, n.id, g.energy[n.id])

    for (a, l) in g.edges.iteritems():
        for b in l:
            print "\\draw[color=red] (%d) -- (%d);" % (a, b)
    print "\\end{tikzpicture}"
    print
    print "\\textbf{Iter: %d, Temperature : %f, dt: %d, Energy : %d, P: %f}" % (k, temp, dt, g.get_energy(), numpy.exp(-float(dt) / temp))
    print
    # print "\\end{figure}"

def show(g):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    for i in g.nodes:
        p = matplotlib.patches.Rectangle((i.x * 10, i.y * 10), 40, 40, facecolor="r")
        ax.add_patch(p)
    l1 = matplotlib.lines.Line2D([0, 1], [0, 1],
           transform=fig.transFigure, figure=fig)
    ax.lines.extend([l1])
    ax.autoscale_view()
    plt.show()

random.seed()

class Composant:
    def __init__(self, id):
        self.id = id
        self.x = (id / 5) * 5
        self.y = (id % 5) * 5

def distance(a, b):
    return abs(a.x - b.x) + abs(a.y - b.y)

class Graph:
    def __init__(self):
        self.edges = dict()
        self.nodes = []

        for i in xrange(25):
            self.nodes.append(Composant(i))
            self.edges[i] = []

        for i in xrange(1, 4):
            for j in xrange(1, 4):
                self.connect(self.nodes[i * 5 + j], self.nodes[i * 5 + j + 1])
                self.connect(self.nodes[i * 5 + j], self.nodes[i * 5 + j - 1])
                self.connect(self.nodes[i * 5 + j], self.nodes[(i+1) * 5 + j])
                self.connect(self.nodes[i * 5 + j], self.nodes[(i-1) * 5 + j])

        for (x, y) in [(0, 1), (0, 5), (4, 3), (4, 9), (20, 21), (15, 20), (24, 23), (24, 19)]:
            self.connect(self.nodes[x], self.nodes[y])

        for (x, y) in [(2, 1), (2, 3), (5, 10), (10, 15), (9, 14), (14, 19), (23, 22), (22, 21)]:
            self.connect(self.nodes[x], self.nodes[y])

        self.compute_energy()

    def shuffle(self):
        # FIXME: doesn't seem great.
        for _ in xrange(25):
            a = self.random_state()
            b = self.random_state()
            a.x, b.x = b.x, a.x
            a.y, b.y = b.y, a.y
        self.compute_energy()

    def connect(self, a, b):
        if b.id in self.edges[a.id]: return
        self.edges[a.id].append(b.id)
        self.edges[b.id].append(a.id)

    def random_state(self):
        nb_nodes = len(self.nodes)
        i = random.randint(0, nb_nodes - 1)
        return self.nodes[i]

    def swap(self, a, b):
        # Update energies
        self.total_energy -= self.energy[a.id] * 2
        self.energy[a.id] = 0
        for i in self.edges[a.id]:
            self.energy[i] -= distance(self.nodes[i], a)
            new_dist = distance(self.nodes[i], b)
            self.energy[i] += new_dist
            self.energy[a.id] += new_dist
            self.total_energy += new_dist * 2

        self.total_energy -= self.energy[b.id] * 2
        self.energy[b.id] = 0
        for i in self.edges[b.id]:
            self.energy[i] -= distance(self.nodes[i], b)
            new_dist = distance(self.nodes[i], a)
            self.energy[i] += new_dist
            self.energy[b.id] += new_dist
            self.total_energy += new_dist * 2

        if a.id in self.edges[b.id]:
            dist = distance(a, b)
            self.total_energy += dist * 2
            self.energy[a.id] += dist * 2 # previously decreased
            self.energy[b.id] += dist

        # Swap positions
        a.x, b.x = b.x, a.x
        a.y, b.y = b.y, a.y

    def energy_if_swap(self, a, b):
        dist = 0
        for i in self.edges[a.id]:
            dist += distance(b, self.nodes[i])
        for i in self.edges[b.id]:
            dist += distance(a, self.nodes[i])

        if a.id in self.edges[b.id]:
            dist += distance(a, b) * 2
        return dist - self.energy[a.id] - self.energy[b.id]

    def compute_energy(self):
        self.energy = dict()
        self.total_energy = 0
        for i in self.nodes:
            self.energy[i.id] = 0
        for (i, l) in self.edges.iteritems():
            for j in l:
                dist = distance(self.nodes[i], self.nodes[j])
                self.energy[i] += dist
                self.total_energy += dist

    def get_energy_safe(self):
        e = 0
        for (i, l) in self.edges.iteritems():
            for j in l:
                e += distance(self.nodes[i], self.nodes[j])
        return e / 2

    def get_energy(self):
        return self.total_energy / 2

def initialize():
    g = Graph()
    g.shuffle()
    return g

def updateTemperature(k):
    return 0.98 * k
    # return numpy.log((1.1 - k) * 100 + 1)

def recuit(g, interface = None):
    current_temp = starting_temp
    k = 0
    nb_swap = 0
    min_energy = 10000
    starting_energy = g.get_energy
    last_gain = 0
    while k < max_iter and g.get_energy() > energy_max:
        e = g.get_energy()
        if e < min_energy:
            min_energy = e
            last_gain = k

        a = g.random_state()
        b = g.random_state()
        if a.id == b.id: continue
        dt = g.energy_if_swap(a, b)
        r = random.random()

        if nb_swap % 300 == 0 or nb_swap % 2500 == 0:
            current_temp = updateTemperature(current_temp)

        # Heat up again when stuck
        if (k - last_gain > max_iter / stall_threshold):
            current_temp += temp_increment
            last_gain = k
            min_energy = e

        #if k > 900 and k < 950:
        #    tikz_show(g, (a, b), temp, dt, k)

        swapped = False
        if r < numpy.exp(-float(dt) / current_temp):
            g.swap(a, b)
            nb_swap += 1
            swapped = True

        k += 1

        if interface:
            interface.update((a.id, b.id), current_temp, dt, k, swapped)

    print k, g.get_energy(), current_temp

if __name__ == "__main__":
    recuit(initialize())
