#! /usr/bin/python

import numpy
import random
import sys
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.Qwt5 import *
import circuit
import copy

scale = 25

class SimulationState(QGraphicsItem):
    def __init__(self, graph, swap, temp, dt, k, swapped):
        super(SimulationState, self).__init__(None)
        self.g = graph
        self.sa = swap[0]
        self.sb = swap[1]
        self.e = graph.get_energy()
        self.t = temp
        self.dt = dt
        self.swapped = swapped

    def paint(self, painter, options, widget = 0):
        s = scale
        g = self.g
        for n1 in range(len(g.edges)):
            for n2 in g.edges[n1]:
                if n1 == self.sa or n2 == self.sa:
                    painter.setPen(QColor(255, 0, 0))
                else:
                    if n1 == self.sb or n2 == self.sb:
                        painter.setPen(QColor(0, 180, 0))
                    else:
                        painter.setPen(QColor(0, 0, 0))
                painter.drawLine(g.nodes[n1].x * s, g.nodes[n1].y * s,
                                 g.nodes[n2].x * s, g.nodes[n2].y * s)
        painter.setPen(QColor(0, 0, 0))
        for n in range(len(g.nodes)):
            if n == self.sa:
                painter.setBrush(QColor(255, 150, 150, 200))
            else:
                if n == self.sb:
                    painter.setBrush(QColor(150, 220, 150, 200))
                else:
                    painter.setBrush(QColor(150, 150, 255, 200))
            x = g.nodes[n].x * s
            y = g.nodes[n].y * s
            r = QRect(x - s * 5 / 3, y - s * 5 / 3,
                      s * 10 / 3, s * 10 / 3)
            painter.drawRect(r)
            painter.drawText(r, Qt.AlignCenter | Qt.AlignVCenter,
                             str(n) + "\nWeight:\n" + str(g.energy[n]))

    def boundingRect(self):
        return QRectF(-50, -50, scale * 20 + 100, scale * 20 + 100)

class View(QGraphicsView):
    def __init__(self, scene):
        super(View, self).__init__(scene)

    def keyPressEvent(self, event):
        k = event.key()
        if k == Qt.Key_Right or k == Qt.Key_Up:
            self.emit(SIGNAL("next()"))
        else:
            if k == Qt.Key_Left or k == Qt.Key_Down:
                self.emit(SIGNAL("prev()"))

class Interface(QWidget):
    def __init__(self):
        super(Interface, self).__init__(None)
        self.graph = None
        self.currentState = 0
        self.states = []
        self.scene = QGraphicsScene()
        self.view = View(self.scene)

        QObject.connect(self.view, SIGNAL("next()"),
                        self.next)
        QObject.connect(self.view, SIGNAL("prev()"),
                        self.prev)

        self.setWindowTitle('Metaheuristiques TP 1 - Recuit simule')
        self.view.centerOn(500, 500)
        self.resize(1024, 768)
        self.slider = QSlider(Qt.Vertical)
        self.slider.setRange(0, 0)
        QObject.connect(self.slider, SIGNAL("valueChanged(int)"),
                        self.display)

        self.progressBar = QProgressBar()
        self.progressBar.setRange(0, 1000)
        self.progressBarLabel = QLabel("Computing simulated annealing")
        self.progressBarLabel.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)

        self.iterationLabel = QLabel("Iteration:")
        self.iterationLabel.setAlignment(Qt.AlignCenter)
        self.iterationEdit = QLineEdit()
        self.iterationEdit.setReadOnly(True)
        self.iterationEdit.setAlignment(Qt.AlignCenter)

        self.temperatureLabel = QLabel("Temperature:")
        self.temperatureLabel.setAlignment(Qt.AlignCenter)
        self.temperatureEdit = QLineEdit()
        self.temperatureEdit.setReadOnly(True)
        self.temperatureEdit.setAlignment(Qt.AlignCenter)

        self.dtLabel = QLabel("dt:")
        self.dtLabel.setAlignment(Qt.AlignCenter)
        self.dtEdit = QLineEdit()
        self.dtEdit.setReadOnly(True)
        self.dtEdit.setAlignment(Qt.AlignCenter)

        self.energyLabel = QLabel("Energy:")
        self.energyLabel.setAlignment(Qt.AlignCenter)
        self.energyEdit = QLineEdit()
        self.energyEdit.setReadOnly(True)
        self.energyEdit.setAlignment(Qt.AlignCenter)

        self.pLabel = QLabel("P:")
        self.pLabel.setAlignment(Qt.AlignCenter)
        self.pEdit = QLineEdit()
        self.pEdit.setReadOnly(True)
        self.pEdit.setAlignment(Qt.AlignCenter)

        self.swapLabel = QLabel("Swap:")
        self.swapLabel.setAlignment(Qt.AlignCenter)
        self.swapEdit = QLineEdit()
        self.swapEdit.setReadOnly(True)
        self.swapEdit.setAlignment(Qt.AlignCenter)

        self.jumpToEdit = QLineEdit()
        self.jumpToEdit.setAlignment(Qt.AlignCenter)
        QObject.connect(self.jumpToEdit, SIGNAL("returnPressed()"),
                        self.jump)
        self.jumpToButton = QPushButton("Jump to")
        QObject.connect(self.jumpToButton, SIGNAL("clicked()"),
                        self.jump)

        self.resetButton = QPushButton("Reset")
        QObject.connect(self.resetButton, SIGNAL("clicked()"),
                        self.reset)

        self.processButton = QPushButton("Process")
        QObject.connect(self.processButton, SIGNAL("clicked()"),
                        self.process)

        self.plot = QwtPlot(QwtText(""))
        self.plot.setAxisTitle(QwtPlot.yLeft, "Temperature")
        self.plot.setAxisTitle(QwtPlot.yRight, "Energy")
        self.plot.enableAxis(QwtPlot.yRight)
        self.plotTimer = QTimer()
        self.plotTimer.setSingleShot(True)
        QObject.connect(self.plotTimer, SIGNAL("timeout()"),
                        self.plot.replot)

        self.plotMarker = QwtPlotMarker()
        self.plotMarker.setLineStyle(QwtPlotMarker.VLine)
        self.plotMarker.setZ(3)

        self.splitter = QSplitter(Qt.Vertical)
        self.splitter.setHandleWidth(10)
        self.splitter.addWidget(self.view)
        self.splitter.addWidget(self.plot)
        self.splitter.setSizes([500, 0])
        lh = QHBoxLayout()
        lv = QVBoxLayout()
        lv.addWidget(self.slider)
        lv3 = QVBoxLayout()
        lv3.addWidget(self.iterationLabel)
        lv3.addWidget(self.iterationEdit)
        lv3.addWidget(self.temperatureLabel)
        lv3.addWidget(self.temperatureEdit)
        lv3.addWidget(self.dtLabel)
        lv3.addWidget(self.dtEdit)
        lv3.addWidget(self.energyLabel)
        lv3.addWidget(self.energyEdit)
        lv3.addWidget(self.pLabel)
        lv3.addWidget(self.pEdit)
        lv3.addWidget(self.swapLabel)
        lv3.addWidget(self.swapEdit)
        lv3.addStretch(1)
        lv3.addWidget(self.resetButton)
        lv3.addWidget(self.processButton)
        lv3.addStretch(1)
        lv3.addWidget(self.jumpToEdit)
        lv3.addWidget(self.jumpToButton)
        lv2 = QVBoxLayout()
        lv2.addWidget(self.splitter)
        lv2.addWidget(self.progressBarLabel)
        lv2.addWidget(self.progressBar)
        lh.addLayout(lv2, 1)
        lh.addLayout(lv3)
        lh.addLayout(lv)
        self.setLayout(lh)

        self.progressBarLabel.hide()
        self.progressBar.hide()

        self.reset()

    def next(self):
        self.display(self.currentState + 1)

    def prev(self):
        self.display(self.currentState - 1)

    def jump(self):
        i = self.jumpToEdit.text().toInt()
        if i[1]:
            self.display(i[0])
        self.jumpToEdit.setText("")

    def update(self, swap, temp, dt, k, s):
        if not self.graph:
            return
        self.states.append(SimulationState(self.graph, swap, temp, dt, k, s))
        self.progressBar.setValue(k * 1000 / circuit.max_iter)

    def reset(self):
        self.graph = circuit.initialize()
        self.states = [SimulationState(self.graph, (-1, -1),
                                       circuit.starting_temp, 0, 0, False)]
        self.iterationEdit.setText("")
        self.temperatureEdit.setText("")
        self.dtEdit.setText("")
        self.energyEdit.setText("")
        self.pEdit.setText("")
        self.swapEdit.setText("")
        self.currentState = 0
        self.slider.setValue(0)
        self.slider.setRange(0, 0)
        self.processButton.setEnabled(True)
        self.display(0)
        self.plotMarker.detach()
        self.plot.clear()
        self.plot.replot()

    def process(self):
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

        self.splitter.hide()
        self.progressBarLabel.show()
        self.progressBar.show()
        self.processButton.setEnabled(False)

        circuit.recuit(self.graph, self)
        self.update((-1, -1), self.states[len(self.states) - 1].t, 0, 0, False)

        t = []
        e = []
        x = []
        i = 0
        sz = len(self.states)
        while i < sz:
            t.append(float(self.states[i].t))
            e.append(float(self.states[i].e))
            x.append(float(i))
            i += 1
        self.plotMarker.detach()
        self.plot.clear()
        ct = QwtPlotCurve("Temperature")
        ct.setPen(QPen(QColor(255, 0, 0)))
        ct.setZ(2)
        ce = QwtPlotCurve("Energy")
        ce.setPen(QPen(QColor(0, 0, 255)))
        ce.setYAxis(QwtPlot.yRight)
        ce.setZ(1)
        ct.setData(x, t)
        ce.setData(x, e)
        ct.attach(self.plot)
        ce.attach(self.plot)
        self.plotMarker.attach(self.plot)

        legend = QwtLegend()
        legend.insert(ct, QLabel("Temperature"))
        legend.insert(ce, QLabel("Energy"))

        self.plot.insertLegend(legend)

        self.plot.replot()

        self.slider.setRange(0, len(self.states) - 1)
        self.currentState = len(self.states) - 1
        self.display(len(self.states) - 1)
        self.splitter.show()
        self.progressBarLabel.hide()
        self.progressBar.hide()

        QApplication.restoreOverrideCursor()

    def display(self, i):
        if not self.graph:
            return
        if i >= len(self.states) or i < 0:
            return
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        for item in self.scene.items():
            self.scene.removeItem(item)
        while self.currentState > i:
            self.currentState -= 1
            cs = self.currentState
            if self.states[cs].swapped:
                self.graph.swap(self.graph.nodes[self.states[cs].sa],
                                self.graph.nodes[self.states[cs].sb])
        while self.currentState < i:
            cs = self.currentState
            if self.states[cs].swapped:
                self.graph.swap(self.graph.nodes[self.states[cs].sa],
                                self.graph.nodes[self.states[cs].sb])
            self.currentState += 1
        s = self.states[i]
        self.scene.addItem(s)
        self.iterationEdit.setText(QString("%1").arg(i))
        self.temperatureEdit.setText(QString("%1").arg(s.t, 0, 'g', 3))
        self.dtEdit.setText(QString("%1").arg(s.dt))
        p = 0.
        if (s.t != 0):
            p = numpy.exp(-float(s.dt) / s.t)
        self.energyEdit.setText(QString("%1").arg(self.graph.get_energy()))
        self.pEdit.setText(QString("%1").arg(p, 0, 'g', 3))
        if self.states[self.currentState].swapped:
            self.swapEdit.setText("Yes")
        else:
            self.swapEdit.setText("No")
        if self.slider.value() != self.currentState:
            self.slider.setValue(self.currentState)
        self.plotMarker.setXValue(self.currentState)
        self.plotTimer.setInterval(100)
        self.plotTimer.start()
        QApplication.restoreOverrideCursor()
        self.scene.invalidate()

app = QApplication(sys.argv)

interface = Interface()
interface.show()

sys.exit(app.exec_())
